

#include <EEPROM.h>

#include "CmdProcessor.h"
#include "MAX31856.h"
#include "LEDDisplayDriver.h"
#include "Average.h"

// Decided not to use this led, because it is inside the box and it will add some heat
//#define USE_STATUS_LED

const byte T1CSpin = 15;
const byte T1SDOpin = 8;  // SDI for Arduino
const byte T2CSpin = 14;
const byte T2SDOpin = 9;  // SDI for Arduino
const byte CLKpin = 16;
const byte SDIpin = 10; // SDO for Arduino

const byte DataPin = A0;
const byte ClockPin = A1;
const byte LoadPin = A2;

LEDDisplayDriver display(DataPin, ClockPin, LoadPin, false, 6);

#define NEOPIXELDDRPORT DDRC
#define NEOPIXELPORT PORTC
#define NEOPIXELBIT PC6

boolean listMode = false;
byte enabled = 3;   // Channel 1 and 2 enabled
unsigned long commTimer = 0;   // Used for turning status blue on communication
unsigned long updateTimer = 0;
unsigned long listTime = 0;
unsigned long listIndex = 0;
char deli = ';';
char comma = ',';
boolean autoMode = true;
byte displayPhase = 0;
byte displayTimer = 1;

float lastTemp1;
float lastCJ1;
byte lastFault1;
float lastTemp2;
float lastCJ2;
byte lastFault2;

CmdProcessor cmdProcessor;
Average avgTC1;
Average avgTC2;
Average avgCJ1;
Average avgCJ2;

Stream *listStream;

char spc[]=" ";

MAX31856 tc1(SDIpin, T1SDOpin, CLKpin, T1CSpin);
MAX31856 tc2(SDIpin, T2SDOpin, CLKpin, T2CSpin);

char eepromId[] = "ThermoSensor";

#ifdef USE_STATUS_LED
//----------------------------------------------------------------------------------------------------
// Simple routine to handle a few NeoPixels, this code only works on a 16MHz Atmel AVR processor
// 10 neopixels takes about 0.3ms or 300us in addition to the initial 50us delay
// Chips: Timing (ns) /T0H/T1H/T0L/T1L/reset frequency byteFormat
// WS2812B /400/800/850/450/50000/ 800kHz GRB 10mA/led
// WS2812 /400/800/850/450/50000/ 800kHz GRB 20mA/led
// PL9823 /350/1360/1360/350/50000/ 585kHz RGB
// TM1804 /400/850/700/320/24000/ 850kHz RGB
// SK6812 /300/600/900/600/60000/ 830kHz GRB
// SK6812RGBW /300/600/900/600/80000/ 830kHz RGBW
// Made by HKJ from lygte-info.dk
void sendNeoPixelData(byte *data, byte len) __attribute__ ((optimize("-O3"))); // Secure fixed optimization, the "Ofast" will unfold the inner loop and spoil the timing
void sendNeoPixelData(byte *data, byte len) {
  bitSet(NEOPIXELDDRPORT, NEOPIXELBIT);       // Set port bit as output
  noInterrupts();                             // If len is more than 100 the millis will probably loose time
  for (byte j = 0; j < len; j++) {
    byte b = *data++;
    // Using a loop for each byte means an error in timing between the bytes, the chips works fine with this.
    for (byte i = 0; i < (byte) 8; i++) {
      if (b & (byte) 0x80) {                // Need 800ns high and 450ns low. The low time is handled by the preparation for next bit/byte
        bitSet( NEOPIXELPORT, NEOPIXELBIT);
        asm volatile(".rept 8\n\tNOP\n\t.endr":::);
        bitClear( NEOPIXELPORT, NEOPIXELBIT);
      } else {                            // Need 400ns (Some chips are nominal 300ns) high and 850ns low. Including some extra low time here.
        bitSet( NEOPIXELPORT, NEOPIXELBIT);
        asm volatile(".rept 2\n\tNOP\n\t.endr":::); // This is the only critical timing
        bitClear( NEOPIXELPORT, NEOPIXELBIT);
        asm volatile(".rept 7\n\tNOP\n\t.endr":::);
      }
      asm volatile(".rept 1\n\tNOP\n\t.endr":::);
      b <<= (byte) 1;
    }
  }
  interrupts();
  // Time for neopixels to latch data, they can do that before next call
  delayMicroseconds(200);
}

//----------------------------------------------------------------------------------------------------
void setStatusLed() {
  byte r = 0;
  byte g = 10;
  byte b = 0;
  if (commTimer >= millis()) {
    b = 40;
  }
  if ((enabled & 1 && lastFault1 != 0) || (enabled & 2 && lastFault2 != 0)) {
    r = 80;
  }

  byte data[3];
  data[0] = g;
  data[1] = r;
  data[2] = b;
  sendNeoPixelData(data, 3);
}
#endif

//----------------------------------------------------------------------------------------------------
// Save current configuration
// This only includes enabled TC, CJ offset and samples setting
void saveConfig() {
  int a = 0;
  EEPROM.put(a, eepromId);

  a = 20;
  EEPROM.put(a, enabled);
  a += sizeof(enabled);
  byte sz = avgTC1.getSize();
  EEPROM.put(a, sz);
  a += sizeof(sz);
  sz = avgTC2.getSize();
  EEPROM.put(a, sz);
  a += sizeof(sz);

  float ofs = tc1.getCJOffset();
  EEPROM.put(a, ofs);
  a += sizeof(ofs);
  ofs = tc2.getCJOffset();
  EEPROM.put(a, ofs);
  a += sizeof(ofs);

}

//----------------------------------------------------------------------------------------------------
// Load saved configuration, this is always done after a reset
void loadConfig() {
  int a = 0;
  char s[20];
  EEPROM.get(a, s);
  if (strcmp(s, eepromId) != 0) return;

  a = 20;
  EEPROM.get(a, enabled);
  a += sizeof(enabled);
  byte sz;
  EEPROM.get(a, sz);
  a += sizeof(sz);
  avgTC1.setSize(sz);
  avgCJ1.setSize(sz);
  EEPROM.get(a, sz);
  a += sizeof(sz);
  avgTC2.setSize(sz);
  avgCJ2.setSize(sz);

  float ofs;
  EEPROM.get(a, ofs);
  a += sizeof(ofs);
  tc1.setCJOffset(ofs);
  EEPROM.get(a, ofs);
  a += sizeof(ofs);
  tc2.setCJOffset(ofs);
}

void clearAverage() {
  avgTC1.clear();
  avgCJ1.clear();
  avgTC2.clear();
  avgCJ2.clear();
}

//----------------------------------------------------------------------------------------------------
void cmdTempQ(Stream *stream) {
  if (enabled & 1) {
    stream->print(lastTemp1, tc1.isVoltMode() ? 3 : 2);
    stream->print(' ');
    stream->print(lastCJ1);
  }
  if ((enabled & 3) == 3) stream->print(' ');
  if (enabled & 2) {
    stream->print(lastTemp2, tc2.isVoltMode() ? 3 : 2);
    stream->print(' ');
    stream->print(lastCJ2);
  }
  stream->println();
}

void streamError(Stream *stream, byte err) {
  if (err == 0) {
    stream->print("- "); return;
  }
  if (err & TC_FAULT_CJ_RANGE) stream->print(F("CJ-R "));
  if (err & TC_FAULT_TC_RANGE) stream->print(F("TC-R "));
  if (err & TC_FAULT_CJ_HIGH) stream->print(F("CJ-H "));
  if (err & TC_FAULT_CJ_LOW) stream->print(F("CJ-L "));
  if (err & TC_FAULT_TC_HIGH) stream->print(F("TC-H "));
  if (err & TC_FAULT_TC_LOW) stream->print(F("TC-L "));
  if (err & TC_FAULT_VOLT) stream->print(F("VOLT "));
  if (err & TC_FAULT_OPEN) stream->print(F("OPEN "));
}

//----------------------------------------------------------------------------------------------------
void cmdErrorQ(Stream * stream) {
  if (enabled & 1) streamError(stream, tc1.getFault());
  if ((enabled & 3) == 3) stream->print(", ");
  if (enabled & 2)streamError(stream, tc2.getFault());
  stream->println();
}

//----------------------------------------------------------------------------------------------------
void cmdEnableQ(Stream *stream) {
  char c = '-';
  if ((enabled & 3) == 3) c = 'B';
  else if (enabled & 1) c = '1';
  else if (enabled & 2) c = '2';
  stream->println(c);
}

//----------------------------------------------------------------------------------------------------
void cmdEnable() {
  char *cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  switch (*cmd) {
    case '0': enabled = 0; break;
    case '1': enabled = 1; break;
    case '2': enabled = 2; break;
    case '3':
    case 'B':
    case 'b':  enabled = 3; break;
  }
}

//----------------------------------------------------------------------------------------------------
void cmdLimitsQ(Stream * stream) {
  if (enabled & 1) {
    stream->print(tc1.readTCRangeLow());
    stream->print(' ');
    stream->print(tc1.readTCRangeHigh());
  }
  if ((enabled & 3) == 3) stream->print(' ');
  if (enabled & 2) {
    stream->print(tc2.readTCRangeLow());
    stream->print(' ');
    stream->print(tc2.readTCRangeHigh());
  }
  stream->println();
}

//----------------------------------------------------------------------------------------------------
void cmdLimits() {
  char *cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  char n = *cmd;
  cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  float low = atof(cmd);
  cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  float high = atof(cmd);
  switch (n) {
    case '1': tc1.setTCRange(low, high); break;
    case '2': tc2.setTCRange(low, high); break;
    case '3':
    case 'b':
    case 'B':
      tc1.setTCRange(low, high);
      tc2.setTCRange(low, high);
      break;
  }
}

//----------------------------------------------------------------------------------------------------
void cmdOffsetQ(Stream * stream) {
  if (enabled & 1) {
    stream->print(tc1.getCJOffset());
  }
  if ((enabled & 3) == 3) stream->print(' ');
  if (enabled & 2) {
    stream->print(tc2.getCJOffset());
  }
  stream->println();
}

//----------------------------------------------------------------------------------------------------
void cmdOffset() {
  char *cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  char n = *cmd;
  cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  float ofs = atof(cmd);
  clearAverage();
  switch (n) {
    case '1': tc1.setCJOffset(ofs); break;
    case '2': tc2.setCJOffset(ofs); break;
    case '3':
    case 'b':
    case 'B':
      tc1.setCJOffset(ofs);
      tc2.setCJOffset(ofs);
      break;
  }
}

//----------------------------------------------------------------------------------------------------
const char sensorTypeName[][4] = {"B", "E", "J", "K", "N", "R", "S", "T", "x8", "x32"};
MaxTCType sensorType[] = {TC_B, TC_E, TC_J, TC_K, TC_N, TC_R, TC_S, TC_T, TC_G8, TC_G32};

//----------------------------------------------------------------------------------------------------
const char *findSensor(MaxTCType s) {
  for (byte i = 0; i < sizeof(sensorType) / sizeof(sensorType[0]); i++) {
    if (sensorType[i] == s) return sensorTypeName[i];
  }
  return "Undef";
}

//----------------------------------------------------------------------------------------------------
void cmdSensorQ(Stream * stream) {
  if (enabled & 1) {
    stream->print(findSensor(tc1.getTCType()));
  }
  if ((enabled & 3) == 3) stream->print(' ');
  if (enabled & 2) {
    stream->print(findSensor(tc2.getTCType()));
  }
  stream->println();
}

//----------------------------------------------------------------------------------------------------
void cmdSensor() {
  char *cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  char n = *cmd;
  cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  MaxTCType t = TC_K;
  for (byte i = 0; i < sizeof(sensorType) / sizeof(sensorType[0]); i++) {
    if (strcasecmp(sensorTypeName[i], cmd) == 0) {
      t = sensorType[i];
      break;
    }
  }
  clearAverage();
  switch (n) {
    case '1': tc1.setTCType(t); break;
    case '2': tc2.setTCType(t); break;
    case '3':
    case 'b':
    case 'B':
      tc1.setTCType(t);
      tc2.setTCType(t);
      break;
  }
}

//----------------------------------------------------------------------------------------------------
void cmdCJTempQ(Stream * stream) {
  if (enabled & 1) {
    stream->print(tc1.readCJTemp());
    stream->print(tc1.isCJAuto() ? " Auto" : " Manual");

  }
  if ((enabled & 3) == 3) stream->print(' ');
  if (enabled & 2) {
    stream->print(tc2.readCJTemp());
    stream->print(tc2.isCJAuto() ? " Auto" : " Manual");
  }
  stream->println();
}

//----------------------------------------------------------------------------------------------------
void cmdCJTemp() {
  char *cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  char n = *cmd;
  cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  if (*cmd == 'a' || *cmd == 'A') {
    switch (n) {
      case '1': tc1.setCJAuto(); break;
      case '2': tc2.setCJAuto(); break;
      case '3':
      case 'b':
      case 'B':
        tc1.setCJAuto();
        tc2.setCJAuto();
        break;
    }
    return;
  }
  float cj = atof(cmd);
  switch (n) {
    case '1': tc1.setCJTemp(cj); break;
    case '2': tc2.setCJTemp(cj); break;
    case '3':
    case 'b':
    case 'B':
      tc1.setCJTemp(cj);
      tc2.setCJTemp(cj);
      break;
  }
}

//----------------------------------------------------------------------------------------------------
byte sampleNo(byte n) {
  byte v = 0;
  while (n > 1) {
    v++;
    n >>= 1;
  }
  return v;
}


//----------------------------------------------------------------------------------------------------
void cmdSamplesQ(Stream * stream) {
  if (enabled & 1) {
    stream->print(tc1.getSamples() + sampleNo(avgTC1.getSize()));
  }
  if ((enabled & 3) == 3) stream->print(' ');
  if (enabled & 2) {
    stream->print(tc2.getSamples() + sampleNo(avgTC2.getSize()));
  }
  stream->println();
}

//----------------------------------------------------------------------------------------------------
void cmdSamples() {
  char *cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  char n = *cmd;
  cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  byte samples = atoi(cmd);
  byte avgSamples = samples > 4 ? 1 << (samples - 4) : 1;
  switch (n) {
    case '1':
      tc1.setSamples(samples);
      avgTC1.setSize(avgSamples);
      avgCJ1.setSize(avgSamples);
      break;
    case '2':
      tc2.setSamples(samples);
      avgTC2.setSize(avgSamples);
      avgCJ2.setSize(avgSamples);
      break;
    case '3':
    case 'b':
    case 'B':
      tc1.setSamples(samples);
      avgTC1.setSize(avgSamples);
      avgCJ1.setSize(avgSamples);
      tc2.setSamples(samples);
      avgTC2.setSize(avgSamples);
      avgCJ2.setSize(avgSamples);
      break;
  }
}

//----------------------------------------------------------------------------------------------------
void cmdModeQ(Stream * stream) {
  stream->println(autoMode ? "Run" : "Sample");
}

//----------------------------------------------------------------------------------------------------
void cmdMode() {
  char *cmd = strtok(NULL, spc);
  if (cmd == NULL) return;
  clearAverage();
  switch (*cmd) {
    case 'R':
    case 'r':
      tc1.setAuto();
      tc2.setAuto();
      autoMode = true;
      break;
    case 'S':
    case 's':
      autoMode = false;
      tc1.start();
      tc2.start();
      break;
  }
}

//----------------------------------------------------------------------------------------------------
void cmdListQ(Stream *stream) {
  listStream = stream;
  char *cmd = strtok(NULL, spc);
  comma = '.';
  deli = '\t';
  if (cmd != NULL) {
    switch (*cmd) {
      case '.': comma = '.'; deli = ','; break;
      case ';':
      case ',': comma = ','; deli = ';'; break;
    }
  }
  listMode = true;
  listTime = millis();
  listIndex = 0;
  listStream->print("Time");
  if (enabled & 1) {
    listStream->print(deli);
    listStream->print("Temp1");
    listStream->print(deli);
    listStream->print("CJ1");
  }
  if (enabled & 2) {
    listStream->print(deli);
    listStream->print("Temp2");
    listStream->print(deli);
    listStream->print("CJ2");
  }
  listStream->println();
}

//----------------------------------------------------------------------------------------------------
void cmdSaveConfig() {
  saveConfig();
}

//----------------------------------------------------------------------------------------------------
PROGMEM const char cmdlist[] = "TEMP?\0ERROR?\0ENABLE?\0ENABLE\0LIMITS?\0LIMITS\0OFFSET?\0OFFSET\0"
                               "SENSOR?\0SENSOR\0CJTEMP?\0CJTEMP\0SAMPLES?\0SAMPLES\0LIST?\0LIST\0"
                               "MODE?\0MODE\0SAVECONFIG\0\0";

//----------------------------------------------------------------------------------------------------
void handleCmds(Stream * stream, char* cmd) {
  listMode = false;
  commTimer = millis() + 1500; // When to turn the blue led off
  switch (cmdidx(cmd, cmdlist)) {
    case 1: cmdTempQ(stream); break;
    case 2: cmdErrorQ(stream); break;
    case 3: cmdEnableQ(stream); break;
    case 4: cmdEnable(); break;
    case 5: cmdLimitsQ(stream); break;
    case 6: cmdLimits(); break;
    case 7: cmdOffsetQ(stream); break;
    case 8: cmdOffset(); break;
    case 9: cmdSensorQ(stream); break;
    case 10: cmdSensor(); break;
    case 11: cmdCJTempQ(stream); break;
    case 12: cmdCJTemp(); break;
    case 13: cmdSamplesQ(stream); break;
    case 14: cmdSamples(); break;
    case 15: cmdListQ(stream); break;
    case 16: cmdListQ(stream); break;
    case 17: cmdModeQ(stream); break;
    case 18: cmdMode(); break;
    case 19: cmdSaveConfig(); break;
  }
}

//----------------------------------------------------------------------------------------------------

void serialPrintComma(float v) {
  if (v < 0) {
    listStream->print('-');
    v = -v;
  }
  long vv = v * 100 + 0.5;
  listStream->print(vv / 100);
  listStream->print(comma);
  int vvv = vv % 100;
  if (vvv < 10) listStream->print('0');
  listStream->print(vvv);
}

//----------------------------------------------------------------------------------------------------
void listCSV() {
  listStream->print(listIndex++);
  if (enabled & 1) {
    listStream->print(deli);
    serialPrintComma(lastTemp1);
    listStream->print(deli);
    serialPrintComma(lastCJ1);
  }
  if (enabled & 2) {
    listStream->print(deli);
    serialPrintComma(lastTemp2);
    listStream->print(deli);
    serialPrintComma(lastCJ2);
  }
  listStream->println();
}


//----------------------------------------------------------------------------------------------------
void updateDisplay() {
  if (--displayTimer > 0) return;
  displayTimer = 7;
  display.clear();
  switch (displayPhase) { // Skip disabled sensors
    case 0: if ((enabled & 1) == 0) displayPhase++; break;
    case 1: if ((enabled & 2) == 0) displayPhase++; break;
  }
  switch (displayPhase) {
    case 0 :
      display.showDigit(digit1, 0);
      if (lastFault1 & TC_FAULT_OPEN)  {
        display.showText("Open", 2, 4);
      } else if (lastFault1 & TC_FAULT_VOLT)  {
        display.showText("Volt", 2, 4);
      } else {
        display.showNum1decimal(lastTemp1, 1, 5);
      }
      break;
    case 1:
      display.showDigit(digit2, 0);
      if (lastFault2 & TC_FAULT_OPEN)  {
        display.showText("Open", 2, 4);
      } else if (lastFault2 & TC_FAULT_VOLT)  {
        display.showText("Volt", 2, 4);
      } else {
        display.showNum1decimal(lastTemp2, 1, 5);
      }
      break;
    case 2:
      if ((lastFault1 | lastFault2) & ~(TC_FAULT_OPEN | TC_FAULT_VOLT)) {
        // Show range faults here, because temperature may still be valid
        display.showText("RANGE");
      } else {
        display.showDigit(digitA, 0);
        // Use average temperature for ambient and subtract 2.5 degree for internal heating
        display.showNum1decimal((lastCJ1 + lastCJ2) / 2 - 2.5, 1, 5);
      }
      break;
  }
  display.update();
  displayPhase = (displayPhase + 1) % 3;
}

//----------------------------------------------------------------------------------------------------
void readValues1() {
  avgTC1.add(tc1.readTCTemp());
  lastTemp1 = avgTC1.get();
  avgCJ1.add(tc1.readCJTemp());
  lastCJ1 = avgCJ1.get();
  lastFault1 = tc1.getFault();
}

//----------------------------------------------------------------------------------------------------
void readValues2() {
  avgTC2.add(tc2.readTCTemp());
  lastTemp2 = avgTC2.get();
  avgCJ2.add(tc2.readCJTemp());
  lastCJ2 = avgCJ2.get();
  lastFault2 = tc2.getFault();
}

//----------------------------------------------------------------------------------------------------
void background() {
  if (updateTimer < millis()) {
    updateTimer = millis() + 200;
    if (autoMode) {
      readValues1();
      readValues2();
    }
    updateDisplay();
  }

  if (!autoMode) {
    if (tc1.isReady()) {
      readValues1();
      tc1.start();
    }
    if (tc2.isReady()) {
      readValues2();
      tc2.start();
    }
  }
#ifdef USE_STATUS_LED
  setStatusLed();
#endif
  if (listMode && listTime < millis()) {
    listTime += 1000;
    listCSV();
  }

}

//----------------------------------------------------------------------------------------------------
void setup() {
  Serial.begin(9600);
  cmdProcessor.init("ThermoSensor,1,1.0", handleCmds);
  tc1.begin();
  tc2.begin();
  tc1.setSamples(TC_SAMPLES_16);
  tc2.setSamples(TC_SAMPLES_16);
  autoMode = false;
  tc1.start();
  tc2.start();
  updateTimer = millis();
  loadConfig();
  // Turn off unused peripherals to save power and thereby heat
  PRR0 = 0x8d;  // Shut down TWI, Timer1, Serial &  ADC
  PRR1 = 0x19;  // Shut down Timer3, Timer4 & USART1
}

//----------------------------------------------------------------------------------------------------
void loop() {  
  cmdProcessor.processInput();
  background();

  // Reduce power consumption by sleeping
  SMCR = 0x01;    // Enable sleep and use the idle mode
  __asm__ __volatile__("\tSLEEP\n");  // Sleep until next millis interrupt  
}
